/*****************************************************************************
 *          DigRUN: a minimal interpreter for educational purpose            *
 ***************************************************************************** 
 *                                                                           *
 * Copyright (c) 2015 Christophe Alias                                       *
 *                                                                           *
 * This program is free software; you can redistribute it and/or modify it   *
 * under the terms of the GNU General Public License as published by the     *
 * Free Software Foundation; either version 2 of the License, or (at your    * 
 * option) any later version.                                                *
 *                                                                           *
 * This program is distributed in the hope that it will be useful, but       *
 * WITHOUT ANY WARRANTY; without even the implied warranty of                *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General  *
 * Public License for more details.                                          *
 *                                                                           *
 * You should have received a copy of the GNU General Public License along   *
 * with this program; if not, write to the Free Software Foundation, Inc.,   *
 * 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.                 *
 *                                                                           *
 * Written by Christophe Alias, Christophe.Alias@ens-lyon.fr                 *
 *                                                                           *
 *****************************************************************************/
#include "interpreter.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

enum op_t { NOP, ADD, SUB, LD, ST, BEQ, LDI, JA, J };

typedef struct {
  enum op_t op;
  int r1;
  int r2;
  int r3;
  int val;
} inst_t;

inst_t code[MAX_INSTRUCTIONS];
int cp; //current instruction

/* Label declaration */

char* label_id[MAX_INSTRUCTIONS]; //id 
unsigned short int label_address[MAX_INSTRUCTIONS]; //address
int current_label;

/* Label use */

char* label_ldi_lo[MAX_INSTRUCTIONS];
char* label_ldi_hi[MAX_INSTRUCTIONS];
char* label_beq[MAX_INSTRUCTIONS];
char* label_j[MAX_INSTRUCTIONS];

void init()
{
  int i;
  cp = -1;
  for(i=0; i<MAX_INSTRUCTIONS; i++)
    {
      code[i].op = NOP;
      label_ldi_lo[i] = strdup("(none)");
      label_ldi_hi[i] = strdup("(none)");
      label_beq[i] = strdup("(none)");
      label_j[i] = strdup("(none)");
    }

  current_label = -1;
}

void next(int* cp)
{
  (*cp)++;
  if((*cp)>=MAX_INSTRUCTIONS)
    {
      printf("Input program is too big\n");
      exit(1);
    }
}

void emit_add(unsigned short int r_target, 
	      unsigned short int r1, 
	      unsigned short int r2)
{
  next(&cp);
  code[cp].op = ADD;
  code[cp].r1 = r_target;
  code[cp].r2 = r1;
  code[cp].r3 = r2;
}

void emit_sub(unsigned short int r_target, 
	      unsigned short int r1, 
	      unsigned short int r2)
{
  next(&cp);
  code[cp].op = SUB;
  code[cp].r1 = r_target;
  code[cp].r2 = r1;
  code[cp].r3 = r2;
}

void emit_ldi(unsigned short int r_target, 
	      unsigned short int value)
{
  next(&cp);
  code[cp].op = LDI;
  code[cp].r1 = r_target;
  code[cp].val = value;

  //  printf("%d %d\n",r_target,value);
}


void emit_ldi_label_lo(unsigned short int r_target, char* label)
{
  next(&cp);
  code[cp].op = LDI;
  code[cp].r1 = r_target;
  label_ldi_lo[cp] = strdup(label);
}

void emit_ldi_label_hi(unsigned short int r_target, char* label)
{
  next(&cp);
  code[cp].op = LDI;
  code[cp].r1 = r_target;
  label_ldi_hi[cp] = strdup(label);
}

void emit_ld(unsigned short int r_target, 
	     unsigned short int r_base, 
	     unsigned short int offset)
{
  next(&cp);
  code[cp].op = LD;
  code[cp].r1 = r_target;
  code[cp].r2 = r_base;
  code[cp].val = offset;
}


void emit_ld_neg(unsigned short int r_target, 
		 unsigned short int r_base, 
		 unsigned short int offset)
{
  next(&cp);
  short int neg = -1*offset;
  code[cp].op = LD;
  code[cp].r1 = r_target;
  code[cp].r2 = r_base;
  code[cp].val = neg;
}


void emit_st(unsigned short int r_source, 
	     unsigned short int r_base, 
	     unsigned short int offset)
{
  next(&cp);
  code[cp].op = ST;
  code[cp].r1 = r_source;
  code[cp].r2 = r_base;
  code[cp].val = offset;
}

void emit_st_neg(unsigned short int r_source, 
		 unsigned short int r_base, 
		 unsigned short int offset)
{
  next(&cp);
  short int neg = -1*offset;
  code[cp].op = ST;
  code[cp].r1 = r_source;
  code[cp].r2 = r_base;
  code[cp].val = neg;
}

void emit_beq(unsigned short int r1, 
	      unsigned short int r2, char* label)
{
  next(&cp);
  code[cp].op = BEQ;
  code[cp].r1 = r1;
  code[cp].r2 = r2;
  label_beq[cp] = strdup(label);
}

void emit_ja(unsigned short int r1, 
	     unsigned short int r2)
{
  next(&cp);
  code[cp].op = JA;
  code[cp].r1 = r1;
  code[cp].r2 = r2;
}

void emit_j(char* label)
{
  next(&cp);
  code[cp].op = J;
  label_j[cp] = strdup(label);
}


void emit_data_num1(unsigned short int value)
{
  next(&cp);
  printf("Warning: data not implemented!\n");
}

void emit_data_num2(unsigned short int value0, unsigned short int value1)
{
  next(&cp);
  printf("Warning: data not implemented!\n");
}

void emit_data_string(char* s)
{
  int i;
  printf("Warning: data not implemented!\n");

  for(i=0; i<strlen(s); i+=2)
      next(&cp);
}

void define_label(char* label)
{
  current_label++;
  label_id[current_label] = strdup(label);
  label_address[current_label] = cp+1;
}

void fix_labels()
{
  int i,j,label_found;
  for(i=0; i<MAX_INSTRUCTIONS; i++)
    //For each instruction using a label
    {
      /*
       * LDI LO(label)
       */
      if(strcmp(label_ldi_lo[i],"(none)") != 0)
	{
	  //Get the label
	  label_found = 0;
	  for(j=0; j<=current_label; j++)
	    if(strcmp(label_id[j],label_ldi_lo[i]) == 0)
	      {
		//Compute the offset
		unsigned short int offset = label_address[j];	    
		//Fix the instruction
		code[i].val = 255 & offset; //Lo part
		label_found = 1;
	      }
	  if(label_found == 0)
	    {
	      printf("Error: undefined label (ldi lo): %s\n",label_ldi_lo[i]);
	      exit(1);
	    }
	}

      /*
       * LDI HI(label)
       */
      if(strcmp(label_ldi_hi[i],"(none)") != 0)
	{
	  //Get the label
	  label_found = 0;
	  for(j=0; j<=current_label; j++)
	    if(strcmp(label_id[j],label_ldi_hi[i]) == 0)
	      {
		//Compute the offset
		unsigned short int offset = label_address[j];	    
		//Fix the instruction
		code[i].val = 255 & (offset>>8); //Hi part
		label_found = 1;
	      }
	  if(label_found == 0)
	    {
	      printf("Error: undefined label (ldi hi): %s\n",label_ldi_hi[i]);
	      exit(1);
	    }
	}

      /*
       * BEQ
       */
      if(strcmp(label_beq[i],"(none)") != 0)
	{
	  //Get the label
	  label_found = 0;
	  for(j=0; j<=current_label; j++)
	    if(strcmp(label_id[j],label_beq[i]) == 0)
	      {
		//Compute the offset
		short int offset = label_address[j] - i - 1;	    

		if(!((offset >= -128) && (offset <= 127)))
		  {
		    printf("Error: ble is too far (delta=%d)\n",offset);
		    exit(1);
		  }

		//Fix the instruction
		code[i].val = offset;
		label_found = 1;
	      }
	  if(label_found == 0)
	    {
	      printf("Error: undefined label (ble): %s\n",label_beq[i]);
	      exit(1);
	    }
	}

      /*
       * J
       */
      if(strcmp(label_j[i],"(none)") != 0)
	//For each instruction using a label (J)
	{
	  //Get the label
	  label_found = 0;
	  for(j=0; j<=current_label; j++)
	    if(strcmp(label_id[j],label_j[i]) == 0)
	      {
		//Compute the offset
		unsigned short int offset = label_address[j];
		//Fix the instruction
		code[i].val = offset;
		label_found = 1;
	      }
	  if(label_found == 0)
	    {
	      printf("Error: undefined label (j): %s\n",label_j[i]);
	      exit(1);
	    }
	}
    }

}	       


//
// Virtual machine itself
//

typedef struct {
  int r[8]; //registers
  int sram8k[8192]; //data ram
} digmips_t;

digmips_t vm; //virtual machine


void print_inst(inst_t inst)
{
  int r1 = inst.r1;
  int r2 = inst.r2;
  int r3 = inst.r3;
  int val= inst.val;

  if(inst.op == NOP)
    printf("nop");
  if(inst.op == ADD)
    printf("add r%d, r%d, r%d",r1,r2,r3);
  if(inst.op == SUB)
    printf("sub r%d, r%d, r%d",r1,r2,r3);
  if(inst.op == LD)
    printf("ld  r%d, [r%d+%d]",r1,r2,val);
  if(inst.op == ST)
    printf("st  r%d, [r%d+%d]",r1,r2,val);
  if(inst.op == BEQ)
    printf("ble r%d, r%d, %d",r1,r2,val);
  if(inst.op == LDI)
    printf("ldi r%d,%d",r1,val);
  if(inst.op == JA)
    printf("ja  r%d, r%d",r1,r2);
  if(inst.op == J)
    printf("j   %d",val);
}

void print_reg(digmips_t vm)
{
  int i;
  for(i=0; i<8; i++)
    printf("r%d: %d\t",i,vm.r[i]);
}

//
// Main routine
//

void exec_code(int debug_mode)
{
  int io = 0;
  int init = 1;

  cp = 0;
  while(cp < MAX_INSTRUCTIONS)
    {
      if(debug_mode)
	{
	  if(!io && !init) { print_reg(vm); printf("\n");}
	  
	  printf("%d:\t",cp);
	  print_inst(code[cp]);
	  printf("\t\t");
	  io = 0;
	  init = 0;
	}
   
      switch(code[cp].op) {
      case NOP: { if(debug_mode) { printf("\n"); return; } break; }
      case ADD: {
	//printf("add\n");

	vm.r[code[cp].r1] = vm.r[code[cp].r2] + vm.r[code[cp].r3];
	break;
      }
      case SUB: {
	//printf("sub\n");

	vm.r[code[cp].r1] = vm.r[code[cp].r2] - vm.r[code[cp].r3];
	break;
      }
      case LD: {
	//printf("ld\n");

	if(code[cp].val == 63)
	  {
	    if(debug_mode) { printf("\n\t==> INPUT:"); fflush(stdout); io = 1;}
	    vm.r[code[cp].r1] = getchar();
	  }
	else
	  vm.r[code[cp].r1] = vm.sram8k[vm.r[code[cp].r2] + code[cp].val];
	break;
      }
      case ST: {
	//printf("st\n");

	if(code[cp].val == 63)
	  {
	    if(debug_mode) { printf("\n\t==> OUTPUT: '"); io = 1;}
	    printf("%c",(char)(vm.r[code[cp].r1]));
	    fflush(stdout);
	    if(debug_mode) printf("'\n");
	  }
	else
	  vm.sram8k[vm.r[code[cp].r2] + code[cp].val] = vm.r[code[cp].r1];
	break;
      }
      case BEQ: {
	//printf("ble\n");

	if(vm.r[code[cp].r1] <= vm.r[code[cp].r2])
	  {
	    cp = cp + code[cp].val + 1;
	    continue;
	  }
	break;
      }
      case LDI: {
	//printf("ldi\n r%d,%d",code[cp].r1,code[cp].val);

	vm.r[code[cp].r1] = code[cp].val;
	break;
      }
      case JA: {
	//printf("ja\n");

	int address = vm.r[code[cp].r1] << 8 | vm.r[code[cp].r2];
	cp = address;
	continue;
      }
      case J:{
	//printf("j\n");
	if(debug_mode) printf("\t");
	if(debug_mode && (code[cp].val == cp)) 
	  {
	    print_reg(vm); printf("\n");
	    return;
	  }

	cp = code[cp].val;
	continue;
      }
      }
      cp++;
    }
}
